/* Copyright 1997 by Jim Aspnes */
/* Distributed under the terms of the GNU General Public License version 2.0 */

/* TEA for Pilot */

#include "ccrypt.h"

/*** NOTE: TEA code is left in only for compatibility with version 1 ***/
/* This version will not encrypt using TEA */

/* (The TEA code and comment below are from */
/*  http://vader.brad.ac.uk/tea/tea.html) */

/************************************************

The Tiny Encryption Algorithm (TEA) by 
David Wheeler and Roger Needham of the
Cambridge Computer Laboratory

Notes:

TEA is a Feistel cipher with XOR and
and addition as the non-linear mixing
functions. 

Takes 64 bits of data in v[0] and v[1].
Takes 128 bits of key in k[0] - k[3].

TEA can be operated in any of the modes
of DES. Cipher Block Chaining is, for example,
simple to implement.

n is the number of iterations. 32 is ample,
16 is sufficient, as few as eight may be OK.
The algorithm achieves good dispersion after
six iterations. The iteration count can be
made variable if required.

Note this is optimised for 32-bit CPUs with
fast shift capabilities. It can very easily
be ported to assembly language on most CPUs.

delta is chosen to be the real part of (the
golden ratio Sqrt(5/4) - 1/2 ~ 0.618034
multiplied by 2^32). 

************************************************/

/* this is here only for backwards compatibility with version 1.0 */
static void
tea_decipher(unsigned long *v, void *kv)
{
   unsigned long *k = (unsigned long *) kv;
   register unsigned long       y=v[0],z=v[1],sum=0xC6EF3720,
				delta=0x9E3779B9,a=k[0],b=k[1],c=k[2],
				d=k[3],n=32;

   /* sum = delta<<5, in general sum = delta * n */

   while(n-->0)
      {
      z -= ((y<<4)+c) ^ (y+sum) ^ ((y>>5)+d);
      y -= ((z<<4)+a) ^ (z+sum) ^ ((z>>5)+b);
      sum -= delta;
      }
   
   v[0]=y; v[1]=z;
}

/* idea.c -- idea encryption
 * 1994 by Risto Paasivirta, paasivir@jyu.fi
 */

#define EXPANDED_IDEA_KEY_LEN (52)

#define INLINE __inline

/* a * b mod 65537
 */
static INLINE unsigned short mul(unsigned short a, unsigned short b);

static INLINE unsigned short mul(unsigned short a, unsigned short b)
{
  long p;
  unsigned long q;

  if (a==0)
    return (unsigned short)(1 - b);
  if (b==0)
    return (unsigned short)(1 - a);
  q = (unsigned long)a * (unsigned long)b;
  if((p = (q & 65535) - (q >> 16)) <= 0)
    p++;
  return (unsigned short)p;
}

/* a^-1 mod 65537
 */

static INLINE unsigned short inv(unsigned short a);

static INLINE unsigned short inv(unsigned short ain)
{
  long a = ain, c = 65537, d, e = 0, f = 1, g;

  while ( a > 0 ) {
    d = c % a;
    g = e - c / a * f;
    c = a; a = d; e = f; f = g;
  }
  if (e < 0)
    e++;
  return (unsigned short)e;
}

/* IDEA core function
 */

static void
Idea(unsigned short *in, unsigned short *out, unsigned short *ks)
{
  unsigned short x0, x1, x2, x3, a, b, i;

  x0 = in[0];
  x1 = in[1];
  x2 = in[2];
  x3 = in[3];
  for (i = 0; i < 8; i++) {
    x0 = mul(*ks++, x0);
    x1 += *ks++;
    x2 += *ks++; 
    x3 = mul(*ks++, x3);
    b = mul(*ks++, x0 ^ x2);
    a = mul(*ks++, b + (x1 ^ x3));
    b += a;
    x0 = a ^ x0;
    x3 = b ^ x3;
    b ^= x1;
    x1 = a ^ x2;
    x2 = b;
  }
  out[0] = mul(*ks++, x0);
  out[1] = *ks++ + x2;
  out[2] = *ks++ + x1;
  out[3] = mul(*ks, x3);
}

/* Expand 128-bit user key to 832-bit encrypt key
 */

static void
ExpandUserKey(unsigned short *key, unsigned short *ks)
{
  unsigned short i;

  for (i = 0; i < 8; i++)
    ks[i] = key[i];
  for (;i < 52; i++) {
    if ((i & 7) == 6) {
      ks[i] = (ks[i - 7] << 9) ^ (ks[i - 14] >> 7);
    } else if ((i & 7) == 7) {
      ks[i] = (ks[i - 15] << 9) ^ (ks[i - 14] >> 7);
    } else {
      ks[i] = (ks[i - 7] << 9) ^ (ks[i - 6] >> 7);
    }
  }
}

/* Invert encrypt key to decrypt key.
 */

static void
InvertIdeaKey(unsigned short *ks, unsigned short *ik)
{
  unsigned short kb[52], i, j;

  for (i = 0, j = 48; i < 52; i += 6, j -= 6) {
    kb[i + 0] = inv(ks[j + 0]);
    if ((i == 0) || (i == 48)) {
      kb[i + 1] = -ks[j + 1];
      kb[i + 2] = -ks[j + 2];
    } else {
      kb[i + 1] = -ks[j + 2];
      kb[i + 2] = -ks[j + 1];
    }
    kb[i + 3] = inv(ks[j + 3]);
    if (i < 48) {
      kb[i + 4] = ks[j - 2];
      kb[i + 5] = ks[j - 1];
    }
  }
  for (i = 0; i < 52; i++)
    ik[i] = kb[i];
}

/* wrapper functions */
static void
idea_in_place(unsigned long *v, void *kv)
{
  unsigned long out[2];

  Idea((unsigned short *) v, (unsigned short *) out, (unsigned short *) kv);
  v[0] = out[0];
  v[1] = out[1];
}

static void
erase_expanded_key(unsigned short *k)
{
  int i;

  for(i = 0; i < EXPANDED_IDEA_KEY_LEN; i++) k[i] = 0;
}

#define ESCAPE (0x1)		/* escape character for hiding nulls */
#define HEADER (0x3)		/* actually 0x1 0x3 */
#define FOOTER (0x4)		/* actually 0x1 0x4 */
#define PAD (8)	/* pad to a multiple of this length before encryption */
#define MAX_PADDED_LEN ((MAXLEN / PAD) * PAD)

/* bracket in with ESCAPE HEADER and ESCAPE FOOTER */
/* escape 0x0 as 0x1 0x1 and 0x1 as 0x1 0x2 */
/* returns new size or ERR_OVERFLOW on overflow */
static int
escape(const struct cbuf *in, struct cbuf *out)
{
  int from;

  out->data[0] = ESCAPE;
  out->data[1] = HEADER;
  out->length = 2;

  for(from=0; from < in->length; from++, out->length++) {
    if (out->length >= MAXLEN - 2) return ERR_OVERFLOW;
    switch(in->data[from]) {
    case 0x0:
    case ESCAPE:
      out->data[out->length++] = ESCAPE;
      if (out->length >= MAXLEN) return ERR_OVERFLOW;
      /* else shift escaped char */
      out->data[out->length] = in->data[from]+1;
      break;
    default: 
      out->data[out->length] = in->data[from];
      break;
    }
  }

  out->data[out->length++] = ESCAPE;
  out->data[out->length++] = FOOTER;
  
  return out->length;
}

/* inverse of above */
/* returns new size or ERR_TRUNCATED if escape occurs as last char */
/* returns ERR_NO_HEADER if it didn't find the header */
/* (e.g. if decrypting a string encrypted by version 1) */
/* Note: ok if in == out */
static int
unescape(const struct cbuf *in, struct cbuf *out)
{
  int from;
  int to;
  int got_header;
  int got_footer;

  got_header = got_footer = 0;

  to = 0;

  for(from = 0; from < in->length; from++) {
    if (in->data[from] == ESCAPE) {

      from++;
      if (from >= in->length) return ERR_TRUNCATED;

      /* else */

      switch(in->data[from]) {

      case HEADER:

	got_header++;
	to = 0;	/* start over */
	break;

      case FOOTER:

	/* we're done */
	got_footer++;
	goto end_of_loop;

      default:
	  out->data[to++] = in->data[from]-1;
      }

    } else {
      /* normal, non-escaped character */
      out->data[to++] = in->data[from];

    }
  }

end_of_loop:

  /* check header and footer */
  if(got_header == 0) {
    return ERR_NO_HEADER;
  } else if(got_footer == 0) {
    return ERR_TRUNCATED;
  } else {
    return out->length = to;
  }
}


/* erases a cbuf */
void
erase(struct cbuf *c)
{
  int i;
  for (i=0; i < MAXLEN; i++) c->data[i] = 0;
  c->length=0;
}

/* copy and null-pad a cbuf */
/* returns new length or ERR_OVERFLOW */
static int
pad(const struct cbuf *in, struct cbuf *out)
{
  int i;

  /* copy the data */
  for(i = 0; i < in->length; i++) out->data[i] = in->data[i];
  out->length = in->length;

  /* hides any trailing nulls */
  
  if (out->length >= MAX_PADDED_LEN) return ERR_OVERFLOW;

  /* else we're ok */
  while(out->length % PAD != 0) out->data[out->length++] = 0;

  return out->length;
}

/* reverses pad */
/* assumes no nulls at end of original! */
static int
unpad(const struct cbuf *in, struct cbuf *out)
{
  int i;

  for(i = 0; i < in->length; i++) {
    out->data[i] = in->data[i];
  }

  out->length = i;

  while(out->length > 0 && out->data[out->length - 1] == '\0') {
    out->length--;
  }

  return out->length;
}

/* pad, encrypt, & escape */
int
code(key k, struct cbuf *v)
{
  struct cbuf padded;
  int ret;
  int i;
  unsigned long last_block[2];
  unsigned long *this_block;
  unsigned short expanded_key[EXPANDED_IDEA_KEY_LEN];

  if((ret=pad(v, &padded)) < 0) {
    erase(&padded);
    return ret;
  }

  /* encrypt using IDEA with CBC */
  ExpandUserKey((unsigned short *) k, expanded_key);

  last_block[0]=0;
  last_block[1]=0;

  for(i=0; i < padded.length; i += PAD) {
    this_block = (unsigned long *) (padded.data + i);

    /* cbc step */
    this_block[0] ^= last_block[0];
    this_block[1] ^= last_block[1];

    idea_in_place(this_block, (void *) expanded_key);

    last_block[0] = this_block[0];
    last_block[1] = this_block[1];
  }

  erase_expanded_key(expanded_key);

  return escape(&padded, v);
}

/* algorithm-independent cipher-block-chaining decryptor */
static void cbc_decrypt(void (*decipher)(unsigned long *, void *),
		        void *k,
			struct cbuf *c)
{
  int i;
  unsigned long last_block[2];
  unsigned long save_block[2];
  unsigned long *this_block;

  /* decrypt using TEA with CBC */
  last_block[0]=0;
  last_block[1]=0;

  for(i=0; i < c->length; i += PAD) {
    this_block = (unsigned long *) (c->data + i);

    /* save encrypted block */
    save_block[0] = this_block[0];
    save_block[1] = this_block[1];

    decipher(this_block, k);

    this_block[0] ^= last_block[0];
    this_block[1] ^= last_block[1];

    last_block[0] = save_block[0];
    last_block[1] = save_block[1];
  }
}

/* reverse of code */
int
uncode(key k, struct cbuf *v)
{
  struct cbuf padded;
  int ret;
  unsigned short expanded_key[EXPANDED_IDEA_KEY_LEN];
  unsigned short inverted_key[EXPANDED_IDEA_KEY_LEN];
  
  /* look for armoring */
  if(((unsigned long *) v->data)[0] == ARMOR_HEADER) {
    unarmor(v, &padded);
    ret = unescape(&padded, &padded);
  } else {
    /* just unescape */
    ret = unescape(v, &padded);
  }

  if(ret == ERR_NO_HEADER) {

    /* assume it's encrypted with TEA */
    cbc_decrypt(tea_decipher, k, &padded);

  } else if(ret < 0) {

    /* some other error */
    return ret;

  } else {

    /* initialize IDEA */
    ExpandUserKey((unsigned short *) k, expanded_key);
    InvertIdeaKey(expanded_key, inverted_key);
    erase_expanded_key(expanded_key);

    /* decrypt IDEA */
    cbc_decrypt(idea_in_place, (void *) inverted_key, &padded);

    /* clear IDEA stuff */
    erase_expanded_key(inverted_key);
  }

  ret = unpad(&padded, v);
  erase(&padded);
  return ret;
}

void
setkey(key k, const char *p, unsigned int len)
{
  unsigned int i;
  int m;			/* multiplier */
  int to;

  /* clear all bits */
  for(i = 0; i < sizeof(key); i++) k[i]=0;

  /* grind through the passphrase */
  /* this hash could be greatly improved */
  /* but it has the advantage that raw 16-byte keys */
  /* are passed through unmodified */
  m = 1;
  to = 0;
  for(i = 0; i < len; i++) {
    k[to++] += m * (*p++);
    if(to >= 16) {
      to = 0;
      m += 6;
    }
  }
}

#ifdef TESTMAIN
int
main(int argc, char **argv)
{
  key k;
  struct cbuf v;
  int i;

  setkey(k, argv[1], strlen(argv[1]));

  v.length = read(0, v.data, MAXLEN);
  code(k, &v);
  write(1, v.data, v.length);
  write(1, "\n", 1);
  uncode(k, &v);
  write(1, v.data, v.length);
  write(1, "\n", 1);

  return 0;
}
#endif

	  
