/* Copyright 1997 by Jim Aspnes */
/* Distributed under the terms of the GNU General Public License version 2.0 */

#include "ccrypt.h"

#define BITS (6)
#define TABLE_LEN (1 << BITS)
#define MASK (TABLE_LEN - 1)

static
const char table[] =
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static
int char2bits(int c)
{
  int i;

  for(i = 0; i < TABLE_LEN; i++) {
    if (c == table[i]) return i;
  }

  return -1;
}

static void
expand_block(const unsigned char in[3], unsigned char out[4])
{
  out[0] = in[0] & MASK;
  out[1] = ((in[0] >> 6) | (in[1] << 2)) & MASK;
  out[2] = ((in[1] >> 4) | (in[2] << 4)) & MASK;
  out[3] = (in[2] >> 2) & MASK;
}

static void
unexpand_block(const unsigned char in[4], unsigned char out[3])
{
  out[0] = in[0] | (in[1] << 6);
  out[1] = (in[1] >> 2) | (in[2] << 4);
  out[2] = (in[2] >> 4) | (in[3] << 2);
}

int
armor(const struct cbuf *in, struct cbuf *out)
{
  int i;
  int j;
  int k;
  unsigned char iblock[3], oblock[4];
  int spacecount;

  /* insert the header */
  ((unsigned long *) out->data)[0] = ARMOR_HEADER;
  out->length = sizeof(unsigned long);

  /* we'll put out a space every SPACING blocks */
  spacecount = 0;

  /* grab input, three tokens at a time */
  for(i = 0; i < in->length; i += 3) {

    /* can we get 3 characters from in? */
    for(j = 0; j < 3 && i+j < in->length; j++) {
      iblock[j] = in->data[i+j];
    }

    /* pad with nulls if we ran out of input */
    for(; j < 3; j++) {
      iblock[j] = '\0';
    }
  
    if(spacecount++ % ARMOR_SPACING == 0) {
      if(out->length >= MAXLEN) {
	return ERR_OVERFLOW;
      } else {
	out->data[out->length++] = ' ';
      }
    }

    /* armor it up */
    if(out->length < MAXLEN - 4) {
      /* expand the bits */
      expand_block(iblock, oblock);
      for(k = 0; k < 4; k++) {
	out->data[out->length++] = table[oblock[k]];
      }
    } else {
      return ERR_OVERFLOW;
    }
  }

  return out->length;
}

int
unarmor(const struct cbuf *in, struct cbuf *out)
{
  int p;
  int fill;
  int b;
  int i;
  unsigned char iblock[4], oblock[3];
  
  fill = 0;

  if(((unsigned long *) in->data)[0] != ARMOR_HEADER) {
    return ERR_NO_HEADER;
  }

  /* else */
  out->length = 0;
  fill = 0;
  for(p = 4; p < in->length; p++) {
    if((b = char2bits(in->data[p])) >= 0) {
      /* got a character */
      iblock[fill++] = b;

      if(fill == 4) {
	/* got a block */
	unexpand_block(iblock, oblock);

	/* tack it onto out */
	for(i = 0; i < 3 && out->length < MAXLEN; i++) {
	  out->data[out->length++] = oblock[i];
	}

	fill = 0;
      }
    }
  }

  /* if(fill != 0) return ERR_TRUNCATED; */

  return out->length;
}

#ifdef TESTMAIN
#include <stdio.h>
void
main(void)
{
  struct cbuf a, b;

  gets(a.data);
  a.length = strlen(a.data);
  armor(&a, &b);
  b.data[b.length] = '\0';
  puts(b.data);
  armor(&b, &a);		/* wipe out what's there */
  unarmor(&b, &a);
  a.data[a.length] = '\0';
  puts(a.data);
}
#endif
